const ngp = require('../ecxngp');
const HashMap = require('hashmap');

const updateStepsModel = {
    getSProf: 'upd_get_sprof',
    updateSProf: 'upd_sprof',
    updateSProfJobs: 'upd_sprof_jobs',
    keyExists: 'upd_keyExists',
    createKey: 'upd_createKey',
    deleteKey: 'upd_deleteKey'
}

const registerStepsModel = {
    isSProf: 'reg_issprof',
    create: 'reg_create',
    keyExists: 'reg_keyExists',
    createKey: 'reg_createKey',
    deleteKey: 'reg_deleteKey'
}


const updateKeySteps = [updateStepsModel.keyExists, updateStepsModel.createKey, updateStepsModel.deleteKey];
const regKeySteps = [registerStepsModel.keyExists, registerStepsModel.createKey, registerStepsModel.deleteKey];

/**
 * Default function for the keyExists step for both storage profile register (POST) and update (PUT) requests
 * @param step
 * @param funcStream
 * @param ecxSessionId
 * @param oFromNg
 * @param cb
 */
function getKeys(step, funcStream, ecxSessionId, oFromNg, cb) {
    const accessKey = oFromNg.spec.subpolicy.find(policy => policy['accesskey'])['accesskey'];

    if(!accessKey.name) {
        // Assign it to accessKeyHREF only when accessKey is a string
        if (typeof accessKey === 'string') {
            storeAccessKeyHref(step, ecxSessionId, accessKey);
            if (funcStream.length > 0) {
                funcStream[0].continue = 'skip';
            }

            step.cb(ecxSessionId, true);
            ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
        }
    }
    else {
        let keyExists = false;
        ngp.ngpCoreHelper.getKeys(ecxSessionId, function(getKeysResObj) {
            const arrayOfKeys = getKeysResObj.rO['keys'];
            const keysLength = arrayOfKeys.length;
            const now = new Date().getTime();
            accessKey.name = accessKey.name + '_' + now;

            for (let i = 0; i < keysLength; i++) {
                if(arrayOfKeys[i]['name'] === accessKey.name) {
                    storeAccessKeyHref(step, ecxSessionId, arrayOfKeys[i]['links']['self']['href']);
                    keyExists = true;
                    break;
                }
            }

            // Skips create key step
            if(keyExists) {
                if ( funcStream.length > 0 ) {
                    funcStream[0].continue = 'skip';
                }

                step.cb(ecxSessionId, true);
                ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
            }
            else {
                step.cb(ecxSessionId, true);
                ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
            }
        });
    }
}

/**
 * Default function for the createKey step for both storage profile register (POST) and update (PUT) requests
 * @param step
 * @param funcStream
 * @param ecxSessionId
 * @param oFromNg
 * @param cb
 */
function createKey(step, funcStream, ecxSessionId, oFromNg, cb) {
    const postBody = oFromNg.spec.subpolicy.find(policy => policy['accesskey'])['accesskey'];

    ngp.sendRequestToEcx(step.verb, step.port, step.path, ecxSessionId, postBody, function(rObj){
        if ( rObj.e ) {
            const failResult = ngp.getGenericEcxErrorResponse(rObj, ecxSessionId, step);
            ngp.logMsg('error', 'Failed to create Access Key ' + postBody.name + ' for session ID=' + ecxSessionId + ' because of error=' + JSON.stringify(e));
            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
            return;
        }

        if ( rObj.response.statusCode != 200 ) {
            const failResult = ngp.getGenericEcxErrorResponse(rObj, ecxSessionId, step);
            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
        }
        else {
            const crO = rObj.rO;

            ngp.logMsg('warn', 'Created Access Key with name=' + postBody.name + ' and href=' + JSON.stringify(crO.links.self.href));
            storeAccessKeyHref(step, ecxSessionId, crO.links.self.href);
            storeCreatedKey(step, ecxSessionId, crO.links.self.href);

            step.cb(ecxSessionId, crO);
            ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
        }
    });
}

/**
 * Default function for the deleteKey step for both storage profile register (POST) and update (PUT) requests
 * @param step
 * @param funcStream
 * @param ecxSessionId
 * @param oFromNg
 * @param cb
 */
function deleteKey(step, funcStream, ecxSessionId, oFromNg, cb) {
    const keyToDelete = getCreatedKey(step, ecxSessionId);

    if (keyToDelete) {
        const urlParse = require('url-parse');
        const path = urlParse(keyToDelete).pathname

        ngp.logMsg('debug', 'Created Access Keys to delete=' + JSON.stringify(keyToDelete));

        ngp.sendRequestToEcx('DELETE', ngp.ecxExternal, path, ecxSessionId, null, function(dlObj){
            if ( dlObj.response.statusCode != 204 ) {
                ngp.logMsg('error', "Deletion of SPP Access Key '" + keyToDelete + "' failed because of error=" + JSON.stringify(dlObj.rO));
            } else {
                step.cb(ecxSessionId, null);
                ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
            }
        });
    }
    else
    {
        step.cb(ecxSessionId, null);
        ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
    }
}

/**
 * Converts POST and PUT request 'subpolicy' to format expected by the core backend.
 * @param sprofTpl
 * @param oFromNg
 * @param ecxSessionId
 * @param updateSprofInfo
 * @param encPassphrase
 */

function convertReplSubpolicies(sprofTpl, oFromNg, ecxSessionId, updateSprofInfo, encPassphrase) {
    const subpolTplHash = ngp.arrayToHash(sprofTpl.spec.subpolicy, 'type');
    let effPolicies = sprofTpl.spec.subpolicy;
    sprofTpl.name = oFromNg.name;
    sprofTpl.description = oFromNg.description;
    sprofTpl.type = oFromNg.type;

    if ( updateSprofInfo )                                                                   // FTI update case, keep description already in the sprof, i.e., GUI won't
        sprofTpl.description = updateSprofInfo.sprofDesc;                                    // alter desc and we don't want value in sprof template to govern ECC-11991


    oFromNg.spec.subpolicy.forEach(function(subpolicy) {
        const createTarget = !!(subpolicy.site && !subpolicy.target);

        if ( subpolicy.type == 'SNAPSHOT' ) {
            const tplTarget = subpolTplHash.SNAPSHOT;
            tplTarget.spec.option.retention = subpolicy.retention;
            tplTarget.participant = true;
            tplTarget.trigger = subpolicy.trigger;
            tplTarget.spec.option.label = subpolicy.label;
        }

        if ( subpolicy.type == 'REPLICATION' && ! subpolicy.software ) {
            const tplTarget = subpolTplHash.REPLICATION;
            tplTarget.trigger = subpolicy.trigger;
            tplTarget.participant = true;
            tplTarget.name = 'Backup Storage Replication';
            tplTarget.description = 'Backup Storage Replication';
            tplTarget.spec.destination = {
                target: createTarget
                    ? {
                        href: subpolicy.site,
                        resourceType: 'site',
                        id: subpolicy.site
                    }
                    : subpolicy.target,
                option: {
                    retention: subpolicy.retention,
                    useEncryption: subpolicy.useEncryption || false
                }
            };
        }

        if ( subpolicy.type == 'REPLICATION' && subpolicy.software ) {
            const tplTarget = subpolTplHash.BACKUP;
            tplTarget.spec.option.retention = subpolicy.retention;
            tplTarget.participant = true;

            delete sprofTpl.spec.option;
            tplTarget.trigger = subpolicy.trigger;                                             // @@ SPROF - per RS and ECC-10979, only BACKUP is target of this trigger

            tplTarget.spec.destination = {
                target: createTarget
                    ? {
                        href: subpolicy.site,
                        resourceType: 'site',
                        id: subpolicy.site
                    }
                    : subpolicy.target,
                option: {
                    useEncryption: subpolicy.useEncryption || false
                }
            };

            if (encPassphrase) {
                tplTarget.spec.option.encPassphrase = {
                    href: encPassphrase
                }
            }
        }

        if ( subpolicy.type === 'SP' ) {
            const tplTarget = subpolTplHash.SP;
            tplTarget.trigger = subpolicy.trigger;

            if ( subpolicy.useVsnap )
                tplTarget.useVsnap = subpolicy.useVsnap;

            if ( subpolicy.destination )
                tplTarget.destination = subpolicy.destination;

            tplTarget.participant = true;
        }

        if ( subpolicy.type === 'SPPOFFLOAD' ) {
            const tplTarget = subpolTplHash.SPPOFFLOAD;
            tplTarget.spec.destination.option.retention = subpolicy.retention;
            tplTarget.trigger = subpolicy.trigger;
            tplTarget.source = subpolicy.source;
            tplTarget.spec.destination.target = subpolicy.target;
            tplTarget.participant = true;
        }

        if ( subpolicy.type === 'SPPARCHIVE' ) {
            const tplTarget = subpolTplHash.SPPARCHIVE;
            tplTarget.spec.destination.option.retention = subpolicy.retention;
            tplTarget.trigger = subpolicy.trigger;
            tplTarget.source = subpolicy.source;
            tplTarget.spec.destination.target = subpolicy.target;
            tplTarget.participant = true;
        }
    });

    for ( let n = effPolicies.length-1; n >= 0; n-- ) {
        if ( ! effPolicies[n].participant )
            effPolicies.splice(n, 1);
        else
            delete effPolicies[n].participant;
    }
}

/**
 * Creates required stream steps.
 * @param requestPayload
 * @param funcSteps
 * @param registrationMode
 * @returns {Array<any>}
 */
function createStepsStream(requestPayload, funcSteps, registrationMode) {
    const _includeKeySteps = includeKeySteps(requestPayload);
    const keySteps = registrationMode ? regKeySteps: updateKeySteps;
    const allSteps = JSON.parse(JSON.stringify(funcSteps));

    if (_includeKeySteps) {
        return allSteps;
    } else {
        return allSteps.filter(el => !keySteps.includes(el.name));
    }
}

/**
 * Checks if steps related to access key should be included.
 * @param requestPayload
 * @returns {boolean}
 */
function includeKeySteps(requestPayload) {
    const slaType = requestPayload['type'];
    const isAccessKeyProvided = requestPayload.spec.subpolicy.some(subpolicy => subpolicy['accesskey']);

    return slaType !== 'snapshot_backup' ? false : isAccessKeyProvided;
}

/**
 * Stores the access key href in the request state
 * @param step
 * @param ecxSessionId
 * @param keyHref
 */
function storeAccessKeyHref(step, ecxSessionId, keyHref) {
    ngp.setRequestState(step.reqId, ecxSessionId, 'accessKeyHREF', keyHref);
}

/**
 * Stores the created key in the request state
 * @param step
 * @param ecxSessionId
 * @param keyHref
 */
function storeCreatedKey(step, ecxSessionId, keyHref) {
    ngp.setRequestState(step.reqId, ecxSessionId, 'createdKeyStorageProfile', keyHref);
}

/**
 * Gets the created key from the request state.
 * @param step
 * @param ecxSessionId
 * @returns {*}
 */
function getCreatedKey(step, ecxSessionId) {
    return ngp.getRequestState(step.reqId, ecxSessionId, 'createdKeyStorageProfile');
}

/**
 * Gets the access key href from the request state.
 * @param step
 * @param ecxSessionId
 * @returns {*}
 */

function getAccessKeyHref(step, ecxSessionId) {
    return ngp.getRequestState(step.reqId, ecxSessionId, 'accessKeyHREF');
}

/**
 * Supports backward compatibility. In case when 'site' is sent instead of 'target', node needs to set proper target values.
 * @param sites
 * @param subPolicies
 */
function setDestinationTarget(sites, subpolicies, updateSprofInfo) {
    const arrayOfSites = sites.rO['sites'];
    const mapSitesNameId = new HashMap();
    const mapSitesNameLinks = new HashMap();

    arrayOfSites.forEach(function (site, i){
        mapSitesNameId.set(arrayOfSites[i]['name'], arrayOfSites[i]['id']);
        mapSitesNameLinks.set(arrayOfSites[i]['name'], arrayOfSites[i].links.self.href);
    });

    subpolicies.forEach(function (subpolicy) {
        if (subpolicy.type === 'REPLICATION' || subpolicy.type === 'BACKUP') {
            subpolicy.spec.destination.target.href = mapSitesNameLinks.get(subpolicy.spec.destination.target.href);
            subpolicy.spec.destination.target.id = mapSitesNameId.get(subpolicy.spec.destination.target.id);
        }

        if(updateSprofInfo && subpolicy.type === 'REPLICATION') {
            subpolicy.id = updateSprofInfo.subpolIDReplication;
        }

        if(updateSprofInfo && subpolicy.type === 'BACKUP') {
            subpolicy.id = updateSprofInfo.subpolIDBackup;
        }
    });
}


// Export constants
exports.updateStepsModel = updateStepsModel;
exports.registerStepsModel = registerStepsModel;

//Export key steps functions
exports.getKeys = getKeys
exports.createKey = createKey;
exports.deleteKey = deleteKey;

//Export functions
exports.createStepsStream = createStepsStream;
exports.storeAccessKeyHref = storeAccessKeyHref;
exports.storeCreatedKey = storeCreatedKey;
exports.getAccessKeyHref = getAccessKeyHref;
exports.getCreatedKey = getCreatedKey;
exports.setDestinationTarget = setDestinationTarget;
exports.convertReplSubpolicies = convertReplSubpolicies;

